/*
                                +---------------------------------+
                                |                                 |
                                |     ***   Rasterizer   ***      |
                                |                                 |
                                |  Copyright   -tHE SWINe- 2002  |
                                |                                 |
                                |           Raster.cpp            |
                                |                                 |
                                +---------------------------------+
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "Load_3ds.h"
#include "Matrix.h"
#include "Pyramid.h"
#include "Font_Eng.h"

#include "Raster.h"

#define PI 3.14159265358979323846263279502884

struct Poly_Edge {                            
    int x, y, num, dxdy;            
};

static int n_Width, n_Height;
// vyska a sirka obrazovky

static float z_delta;
static Pyramid pyramid;
// klipovaci pyramida

static Poly_Edge left_edge_buff[7];
static Poly_Edge right_edge_buff[7];
static Poly_Edge *p_left_edge;
static Poly_Edge *p_right_edge;
static int n_left_edge_num, n_right_edge_num;
// edge

int Init_Engine(int n_New_Width, int n_New_Height)
{
    n_Width = n_New_Width;
    n_Height = n_New_Height;
    // nastav velikost obrazu

    z_delta = (float)fabs((float)n_Width / ((float)tan(90 * (PI / 180) / 2) * 2));
    // perspektivni zkresleni pro hel 90

    SetFOV((float)n_Width - 4, (float)n_Height - 4, z_delta, &pyramid);
    // klipovaci pyramida (je o trochu men kvli nepesnostem)

    return 1;
}

void Terminate_Engine()
{
}

int Add_Edge(Vertex *v_cur, Vertex *v_next, Poly_Edge *edge)
{
    float y1, y2;
    int num;

    y1 = (float)ceil(v_cur->y);
    y2 = (float)ceil(v_next->y);
    num = (int)(y2 - y1);

    if(!num)
        return 0;

    edge->dxdy = (int)((v_next->x - v_cur->x) / (v_next->y - v_cur->y) * 65536);
    edge->x = (int)(v_cur->x * 65536 + (y1 - v_cur->y) * edge->dxdy);
    edge->y = (int)y1;
    edge->num = num;
    // spot prvn x a y, sklon (dxdy) a poet obrazovch dek

    return 1;
}

int Create_Edges(Polygon *p_poly)
{
    int n_bottom, n_top, n_cur;
    float f_min_y;
    float f_max_y;

    n_left_edge_num = 0;
    n_right_edge_num = 0;
    // poty hran

    f_min_y =  0x10000;
    f_max_y = -0x10000;
    n_cur = 0;
    while(n_cur < p_poly->n_vertex_num) {
        if(p_poly->vertex[n_cur].y < f_min_y) {
            f_min_y = p_poly->vertex[n_cur].y;
            n_top = n_cur;
        }
        if(p_poly->vertex[n_cur].y > f_max_y) {
            f_max_y = p_poly->vertex[n_cur].y;
            n_bottom = n_cur;
        }
        n_cur ++;
    }
    // najde nejvrchnj a nejspodnj vrchol

    if((int)ceil(f_min_y) >= (int)ceil(f_max_y))
        return 0;
    // zkontroluje zda se vejde do rastru

    n_cur = n_top;
    while(n_cur != n_bottom) {
        if(Add_Edge(&p_poly->vertex[n_cur], &p_poly->vertex[(n_cur + 1) %
           p_poly->n_vertex_num], &right_edge_buff[n_right_edge_num]))
            n_right_edge_num ++;
        n_cur = (n_cur + 1) % p_poly->n_vertex_num;
    }
    // sjede pravou vtev

    n_cur = n_bottom;
    while(n_cur != n_top) {
        if(Add_Edge(&p_poly->vertex[(n_cur + 1) % p_poly->n_vertex_num],
           &p_poly->vertex[n_cur], &left_edge_buff[n_left_edge_num]))
            n_left_edge_num ++;
        n_cur = (n_cur + 1) % p_poly->n_vertex_num;
    }

    p_left_edge = left_edge_buff + n_left_edge_num - 1;
    p_right_edge = right_edge_buff;

    return 1;
}

int DrawPolygon(Polygon *p, unsigned __int32 *video, unsigned __int32 color)
{
    int start_y, left_x, right_x;
    int l, c;

    if(!Create_Edges(p)) 
        return 0;

    start_y = p_right_edge->y;
    left_x = p_left_edge->x;
    right_x = p_right_edge->x;
    // edge properties

    video += n_Width * start_y;
    // set pointer to first scanline ...

    while(1) {
        c = left_x >> 16;
        l = right_x >> 16;
        // back to int

        for(; l < c; l ++)
            video[l] = color;
        // fill horizontal segment

        video += n_Width;
        // next line ...

        if(-- p_left_edge->num == 0) {
            if(p_left_edge -- == left_edge_buff)
                return 1;
            // end of buffer - polygon was drawn
            left_x = p_left_edge->x;
        } else
            left_x += p_left_edge->dxdy;

        if(-- p_right_edge->num == 0) {
            p_right_edge ++;
            right_x = p_right_edge->x;
        } else 
            right_x += p_right_edge->dxdy;
        // process edges ...
    }

    return 1;
}

void DrawObject(Matrix *m_object, Matrix *m_camera,
    Object *p_object, unsigned __int32 *video, unsigned __int32 color)
{
    unsigned __int32 n_color;
    float tx, ty, tz;
    float x, y, z;    
    Matrix m, im;
    Polygon poly;
    Vertex v[3];
    Face *cur_face, face;
    int i, j;

    face.vertex[0] = &v[0];
    face.vertex[1] = &v[1];
    face.vertex[2] = &v[2];
    // nastav vertexy pro nai doasnou face

    Matrix_Inverse(m_camera, &m);
    Matrix_Multiply(&m, m_object, &m);
    Matrix_Inverse(&m, &im);
    // spot transforman matice

    for(i = 0; i < p_object->n_face_num; i ++){
        if((p_object->face[i].normal.c * im[3][2] +
           p_object->face[i].normal.a * im[3][0] +
           p_object->face[i].normal.b * im[3][1] +
           p_object->face[i].normal.d) > 0.1)
            continue;
        // byckface culling

        cur_face = &p_object->face[i];
        for(j = 0; j < 3; j ++){
            x = cur_face->vertex[j]->x;
            y = cur_face->vertex[j]->y;
            z = cur_face->vertex[j]->z;
            //
            tx = x * m[0][0] + y * m[1][0] + 
                 z * m[2][0] +     m[3][0];
            ty = x * m[0][1] + y * m[1][1] + 
                 z * m[2][1] +     m[3][1];
            tz = x * m[0][2] + y * m[1][2] + 
                 z * m[2][2] +     m[3][2];
            //
            face.vertex[j]->x = tx;
            face.vertex[j]->y = ty;
            face.vertex[j]->z = tz;
        }

        poly = Clip_Face(&face, &pyramid);
        // proene face pyramidou

        x = p_object->face[i].normal.a;
        y = p_object->face[i].normal.b;
        z = p_object->face[i].normal.c;
        tx = x * m[0][0] + y * m[1][0] + z * m[2][0];
        ty = x * m[0][1] + y * m[1][1] + z * m[2][1];
        tz = x * m[0][2] + y * m[1][2] + z * m[2][2];
        // natransformuje normlu

        x = (float)sqrt(tx * tx + ty * ty + tz * tz);
        // normalizuje ji

        j = abs((int)(tz / x * 255));
        if(j > 255)
            j = 255;
        // a spot osvtlen podle jednoduchho pravidla
        // m je face piklonnj ke kamee, tm vc bude
        // osvtlen. to vlastn znamen m vt bude z-ov
        // sloka normlovho vektoru vt (nebo x a y men)
        
        n_color = ((((color & 0xff0000) * j) & 0xff000000) +
                   (((color & 0x00ff00) * j) & 0x00ff0000) +
                   (((color & 0x0000ff) * j) & 0x0000ff00)) >> 8;
        // spot novou barvu - rozlo na sloky R, B a B
        // a ty vynsob svtlost

        if(poly.n_vertex_num) { // pokud je face vidt, bude to nenulov slo
            for(j = 0; j < poly.n_vertex_num; j ++) {
                if(poly.vertex[j].z < 0.1) {
                    poly.n_vertex_num = 0;
                    break;
                }
                // pokud je z - souadnice moc mal, zahodme polygon

                poly.vertex[j].x = (n_Width / 2) + (poly.vertex[j].x *
                    z_delta) / poly.vertex[j].z;
                poly.vertex[j].y = (n_Height / 2) + (poly.vertex[j].y *
                    z_delta) / poly.vertex[j].z;
                // perspektivn korekce
            }

            DrawPolygon(&poly, video, n_color);
            // vykresl polygon na obrazovku
        }
    }
}
